/*
 *  hnv_cone.c
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/15/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 *  2007-10-06 code tuning of remove_redundant_peaks()
 */

#include <stdio.h>
#include <stdlib.h>
#include "hnv_main.h"
#include "hnv_computation_main.h"
#include "hnv_tetrahedron.h"


struct hnv_peak {
	HNV_PEAK_TYPE redundance;	// HNV_REDUNDANT, HNV_NOT_REDUNDANT, or HNV_UNKNOWN
	struct hnv_coor_4d coor;
	int sum_mod3;		// k+l+m+n  mod (dim_of_lattice - 1)
};

struct hnv_cone {
	int dim_of_lattice;
	int peak_ct;
	int peak_size;
	struct hnv_peak * peaks;
};

//===============================================
// Private data
//===============================================

//===============================================
// Public data
//===============================================

struct hnv_cone * my_standard_cone;
struct hnv_cone * my_standard_roof;
struct hnv_cone * my_conjugate_roof;


//===============================================
// Public methods  (part 1)
//===============================================

void print_hnv_cone_short(struct hnv_cone * std_cone)
{
	int p_ct;
	
	fprintf(STDERR, "print_hnv_cone_short() : START of cone\n");
	fprintf(STDERR, "print_hnv_cone_short() : peak_ct = %d, peak_size = %d, dim_of_lattice = %d\n",
		std_cone->peak_ct, std_cone->peak_size, std_cone->dim_of_lattice);
	if (std_cone->peaks != NULL) {
		for (p_ct = 0; p_ct < std_cone->peak_size; p_ct++) {
			if (std_cone->peaks[p_ct].redundance != HNV_REDUNDANT) {
				// print_hnv_peak(std_cone->peaks + p_ct);
				fprintf(STDERR, "print_hnv_cone_short() : coor[%d] = { %8.3f, %8.3f, %8.3f, %8.3f }, ", 
					p_ct, std_cone->peaks[p_ct].coor.k, std_cone->peaks[p_ct].coor.l, 
					std_cone->peaks[p_ct].coor.m, std_cone->peaks[p_ct].coor.n);
				if (std_cone->peaks[p_ct].sum_mod3 >= 0)
					fprintf(STDERR, " sum mod = %d, ", std_cone->peaks[p_ct].sum_mod3);
				else
					fprintf(STDERR, " sum mod = -, ");
				switch (std_cone->peaks[p_ct].redundance) {
					case HNV_REDUNDANT:
						fprintf(STDOUT, "redundance = REDUNDANT\n");
						break;
					case HNV_NOT_REDUNDANT:
						fprintf(STDOUT, "redundance = NOT_REDUNDANT\n");
						break;
					case HNV_UNKNOWN_P:
						fprintf(STDOUT, "redundance = UNKNOWN\n");
						break;
					case HNV_INVALID_P:
						fprintf(STDOUT, "redundance = INVALID\n");
						break;
					default:
						fprintf(STDOUT, "ERROR redundance = invalid value\n");
						break;
				}
			}
		}
	}
	fprintf(STDERR, "print_hnv_cone_short() : END of cone\n");
	return;
}


void initialize_hnv_cone(struct hnv_cone * std_cone)
{
	std_cone->dim_of_lattice = 0;
	std_cone->peak_ct = 0;
	std_cone->peak_size = 0;
	std_cone->peaks = NULL;
	return;
}


struct hnv_cone * allocate_hnv_cone(int dim_ct, int peak_ct)
{
	struct hnv_cone * standard_cone;
	int p_ct;
	
	standard_cone = (struct hnv_cone *) malloc(sizeof(struct hnv_cone));
	if (standard_cone == NULL) {
		fprintf(STDERR, "allocate_hnv_cone() : ERROR allocation failed (cone)\n");
		return (struct hnv_cone *) NULL;
	}
	initialize_hnv_cone(standard_cone);
	standard_cone->dim_of_lattice = dim_ct;
	standard_cone->peak_ct = 0;
	standard_cone->peak_size = peak_ct;
	standard_cone->peaks = (struct hnv_peak *) malloc(peak_ct * sizeof(struct hnv_peak));
	if (standard_cone->peaks == NULL) {
		fprintf(STDERR, "allocate_hnv_cone() : ERROR allocation failed (cone->peaks[%d])\n", peak_ct);
		free(standard_cone);
		return (struct hnv_cone *) NULL;
	}
	for (p_ct = 0; p_ct < peak_ct; p_ct++) {
		standard_cone->peaks[p_ct].redundance = HNV_INVALID_P;
		standard_cone->peaks[p_ct].coor.k = 0.0;
		standard_cone->peaks[p_ct].coor.l = 0.0;
		standard_cone->peaks[p_ct].coor.m = 0.0;
		standard_cone->peaks[p_ct].coor.n = 0.0;
		standard_cone->peaks[p_ct].sum_mod3 = -1;
	}
	return standard_cone;
}


void free_hnv_cone(struct hnv_cone * cone)
{
	if (cone->peaks != NULL)
		free(cone->peaks);
	if (cone != NULL)
		free(cone);
	return;
	
}


//===============================================
// Private methods
//===============================================
//
// void print_hnv_peak(struct hnv_peak * cur_peak);
// void load_peak_coordinate(struct hnv_peak * cur_peak, char * seq_line);
// struct hnv_cone * compute_conjugate_cone(struct hnv_cone * std_cone);
// struct hnv_cone * compute_standard_cone(struct hnv_cone * conj_roof);
// void copy_hnv_cone_compressed(struct hnv_cone * tgt_cone, struct hnv_cone * src_cone);
// float third_smallest(float value0, float value1, float value2, float value3);
// void compute_roof_for_every_four_peaks(struct hnv_cone * std_roof,
//				          struct hnv_cone * std_cone);
// void initialize_hnv_cone(struct hnv_cone * std_cone);
// int str_to_float(float * val, char * seq_line);
// struct hnv_peak * get_cone_peaks(struct hnv_cone * std_cone);
// struct hnv_cone * extract_sub_cone(struct hnv_cone * std_cone, int mod_value);
// struct hnv_cone * merge_three_cones(struct hnv_cone * work_cone0, struct hnv_cone * work_cone1,
//				       struct hnv_cone * work_cone2)
// struct hnv_cone * compute_roof_sub(struct hnv_cone * std_cone);   ==> See Public part
//-----------------------------------------------

void print_hnv_peak(struct hnv_peak * cur_peak)
{
	fprintf(STDERR, "print_hnv_peak() : coor = { %8.3f, %8.3f, %8.3f, %8.3f }, ", 
		cur_peak->coor.k, cur_peak->coor.l, cur_peak->coor.m, cur_peak->coor.n);
	if (cur_peak->sum_mod3 >= 0)
		fprintf(STDERR, " sum mod = %d, ", cur_peak->sum_mod3);
	else
		fprintf(STDERR, " sum mod = -, ");
	switch (cur_peak->redundance) {
		case HNV_REDUNDANT:
			fprintf(STDOUT, "redundance = REDUNDANT\n");
			break;
		case HNV_NOT_REDUNDANT:
			fprintf(STDOUT, "redundance = NOT_REDUNDANT\n");
			break;
		case HNV_UNKNOWN_P:
			fprintf(STDOUT, "redundance = UNKNOWN\n");
			break;
		case HNV_INVALID_P:
			fprintf(STDOUT, "redundance = INVALID\n");
			break;
		default:
			fprintf(STDOUT, "ERROR redundance = invalid value\n");
			break;
	}
	return;
}


//  seq_line = " - 12," " 68 ", " +7}", ... 
int str_to_float(float * val, char * seq_line)
{
	int seq_ind;
	int work;
	float sign = 1.0;
	
	work = 0;
	seq_ind = 0;
	while (seq_line[seq_ind] == ' ') seq_ind++;
	if (seq_line[seq_ind] == '-') {
		sign = -1.0;
		seq_ind++;
	} else if (seq_line[seq_ind] == '+') {
		sign = 1.0;
		seq_ind++;
	}
	while (seq_line[seq_ind] == ' ') seq_ind++;
	while ((seq_line[seq_ind] != ' ') && (seq_line[seq_ind] != ',') 
	       && (seq_line[seq_ind] != '}')) {
		if ((seq_line[seq_ind] -'0' < 0) || ('9' - seq_line[seq_ind] < 0)) {
			fprintf(STDERR, "str_to_float() : ERROR seq_line[%d]='%.1s'\n",
				seq_ind, seq_line + seq_ind);
			exit(1);
		}
		work = work * 10 + (seq_line[seq_ind] -'0');
		seq_ind++;
	}
	work = sign * work;
	while (seq_line[seq_ind] == ' ') seq_ind++;
	if (seq_line[seq_ind] == ',') seq_ind++;
	*val = (float) work;
	return seq_ind;
}


//  seq_line = "{ x0, y0, z0, w0 }" 
void load_peak_coordinate(struct hnv_peak * cur_peak, char * seq_line)
{
	int seq_ind;
	
	seq_ind = 1;	// skip '{'
	seq_ind = seq_ind + str_to_float(&(cur_peak->coor.k), seq_line + seq_ind);
	seq_ind = seq_ind + str_to_float(&(cur_peak->coor.l), seq_line + seq_ind);
	seq_ind = seq_ind + str_to_float(&(cur_peak->coor.m), seq_line + seq_ind);
	seq_ind = seq_ind + str_to_float(&(cur_peak->coor.n), seq_line + seq_ind);
	return;
}


void copy_hnv_cone_compressed(struct hnv_cone * tgt_cone, struct hnv_cone * src_cone)
{
	int p_ct;
	int array_ct;
	
	if (tgt_cone->peak_size < src_cone->peak_ct) {
		fprintf(STDERR, 
			"copy_hnv_cone_compressed() : ERROR tgt_cone->peak_size=%d < src_cone->peak_size=%d\n",
			tgt_cone->peak_size, src_cone->peak_size);
		exit(1);
	}
	tgt_cone->dim_of_lattice = src_cone->dim_of_lattice;
	tgt_cone->peak_ct = src_cone->peak_ct;
	p_ct = 0;
	for (array_ct = 0; array_ct < src_cone->peak_size; array_ct++) {
		if (src_cone->peaks[array_ct].redundance == HNV_REDUNDANT)
			continue;
		tgt_cone->peaks[p_ct].redundance = src_cone->peaks[array_ct].redundance;
		tgt_cone->peaks[p_ct].coor.k = src_cone->peaks[array_ct].coor.k;
		tgt_cone->peaks[p_ct].coor.l = src_cone->peaks[array_ct].coor.l;
		tgt_cone->peaks[p_ct].coor.m = src_cone->peaks[array_ct].coor.m;
		tgt_cone->peaks[p_ct].coor.n = src_cone->peaks[array_ct].coor.n;
		tgt_cone->peaks[p_ct].sum_mod3 = src_cone->peaks[array_ct].sum_mod3;
		p_ct++;
	}
	if(p_ct != tgt_cone->peak_ct) {
		fprintf(STDERR, 
			"copy_hnv_cone_compressed() : ERROR p_ct(=%d) != tgt_cone->peak_ct(=%d), tgt_cone->peak_size=%d\n",
			p_ct, tgt_cone->peak_ct, tgt_cone->peak_size);
		exit(1);
	}
	return;
}


float third_smallest(float value0, float value1, float value2, float value3)
{
	int index[4] ={0, 1, 2, 3};
	float val[4];
	float min;
	int left, right, save_left;
	int cur_min_index;
	
	val[0]=value0;
	val[1]=value1;
	val[2]=value2;
	val[3]=value3;
	
	for (left = 0; left < 3; left++) {
		min = val[index[left]];
		cur_min_index = left;
		for (right = left + 1; right < 4; right++) {
			if (min > val[index[right]]) {
				min = val[index[right]];
				cur_min_index = right;
			}
		}
		save_left = index[left];
		index[left] = index[cur_min_index];
		index[cur_min_index] = save_left;
	}
	
	return val[index[2]];
}
					
								 
void compute_roof_for_every_four_peaks(struct hnv_cone * std_roof,
				       struct hnv_cone * std_cone)
{
	int roof_index;
	int array_size;
	int p0, p1, p2, p3;
	
	copy_hnv_cone_compressed(std_roof, std_cone);
	roof_index = std_roof->peak_ct;
	array_size = std_cone->peak_size;
	if (array_size < 4) {
		fprintf(STDERR, 
			"compute_roof_for_every_four_peaks() : ERROR std_cone->peak_size=%d\n",
			std_cone->peak_size);
		exit(1);
	}
	for (p0 = 0; p0 < array_size; p0++) {
		if (std_cone->peaks[p0].redundance == HNV_REDUNDANT)
			continue;
		std_cone->peaks[p0].redundance = HNV_REDUNDANT;
		for (p1 = p0 + 1; p1 < array_size; p1++) {
			if (std_cone->peaks[p1].redundance == HNV_REDUNDANT)
				continue;
			std_cone->peaks[p1].redundance = HNV_REDUNDANT;
			for (p2 = p1 + 1; p2 < array_size; p2++) {
				if (0)
					fprintf(STDOUT, 
"compute_roof_for_every_four_peaks() : DEBUG  roof_index=%d, (p0, p1, p2) = (%d, %d, %d), cur_peak_ct = %d\n",
						roof_index, p0, p1, p2, array_size);
				if (std_cone->peaks[p2].redundance == HNV_REDUNDANT)
					continue;
				std_cone->peaks[p2].redundance = HNV_REDUNDANT;
				for (p3 = p2 + 1; p3 < array_size; p3++) {
					if (0)
						fprintf(STDOUT,
"compute_roof_for_every_four_peaks() : DEBUG  roof_index=%d, (p0, p1, p2, p3) = (%d, %d, %d, %d), cur_peak_ct = %d\n",
							roof_index, p0, p1, p2, p3, array_size);
					if (std_cone->peaks[p3].redundance == HNV_REDUNDANT)
						continue;
					std_cone->peaks[p3].redundance = HNV_REDUNDANT;
					
					if (roof_index >= std_roof->peak_size) {
						fprintf(STDERR, 
"compute_roof_for_every_four_peaks() : ERROR overflow (roof_index=%d < src_roof->peak_size=%d)\n",
							roof_index, std_roof->peak_size);
						exit(1);
					}
					
					// compute roof
					std_roof->peaks[roof_index].coor.k 
						= third_smallest(std_cone->peaks[p0].coor.k,
								 std_cone->peaks[p1].coor.k,
								 std_cone->peaks[p2].coor.k,
								 std_cone->peaks[p3].coor.k);
					std_roof->peaks[roof_index].coor.l
						= third_smallest(std_cone->peaks[p0].coor.l,
								 std_cone->peaks[p1].coor.l,
								 std_cone->peaks[p2].coor.l,
								 std_cone->peaks[p3].coor.l);
					std_roof->peaks[roof_index].coor.m 
						= third_smallest(std_cone->peaks[p0].coor.m,
								 std_cone->peaks[p1].coor.m,
								 std_cone->peaks[p2].coor.m,
								 std_cone->peaks[p3].coor.m);
					std_roof->peaks[roof_index].coor.n 
						= third_smallest(std_cone->peaks[p0].coor.n,
								 std_cone->peaks[p1].coor.n,
								 std_cone->peaks[p2].coor.n,
								 std_cone->peaks[p3].coor.n);
					std_roof->peaks[roof_index].redundance = HNV_UNKNOWN_P;
					std_roof->peaks[roof_index].sum_mod3 = -1;
					roof_index++;
					std_roof->peak_ct++;
								 
					std_cone->peaks[p3].redundance = HNV_NOT_REDUNDANT;
					std_roof->peaks[p3].redundance = HNV_UNKNOWN_P;
				}
				std_cone->peaks[p2].redundance = HNV_NOT_REDUNDANT;
				std_roof->peaks[p2].redundance = HNV_UNKNOWN_P;
			}
			std_cone->peaks[p1].redundance = HNV_NOT_REDUNDANT;
			std_roof->peaks[p1].redundance = HNV_UNKNOWN_P;
		}
		std_cone->peaks[p0].redundance = HNV_NOT_REDUNDANT;
		std_roof->peaks[p0].redundance = HNV_UNKNOWN_P;
	}
	if (0) {
		int p_ct;
		int array_ct;
		
		p_ct = 0;
		for (array_ct = 0 ; array_ct < std_roof->peak_size; array_ct++) {
			if (std_roof->peaks[array_ct].redundance == HNV_REDUNDANT)
				continue;
			p_ct++;
		}
		if(p_ct != std_roof->peak_ct) {
			fprintf(STDERR, 
				"compute_roof_for_every_four_peaks() : ERROR p_ct(=%d) != std_roof->peak_ct(=%d), std_roof->peak_size=%d\n",
				p_ct, std_roof->peak_ct, std_roof->peak_size);
			exit(1);
		} else {
			fprintf(STDOUT, 
				"compute_roof_for_every_four_peaks() : DEBUG p_ct(=%d) == std_roof->peak_ct(=%d), std_roof->peak_size=%d\n",
				p_ct, std_roof->peak_ct, std_roof->peak_size);
		}
	}
	return;
}


struct hnv_peak * get_cone_peaks(struct hnv_cone * std_cone)
{
	return std_cone->peaks;
}


struct hnv_cone * extract_sub_cone(struct hnv_cone * std_cone, int mod_value)
{
	struct hnv_cone * sub_cone;
	int p_ct;
	int array_ct;
	
	// Count peaks
	p_ct = 0;
	for (array_ct = 0 ; array_ct < std_cone->peak_size; array_ct++) {
		if (std_cone->peaks[array_ct].redundance == HNV_REDUNDANT)
			continue;
		if (std_cone->peaks[array_ct].sum_mod3 != mod_value)
			continue;
		p_ct++;
	}
	sub_cone = allocate_hnv_cone(std_cone->dim_of_lattice, p_ct);
	if (sub_cone == NULL) {
		fprintf(STDERR, "extract_sub_cone() : ERROR allocation failed (sub_cone[%d])\n",
			p_ct);
		exit(1);
	}
	
	// Copy peaks
	p_ct = 0;
	for (array_ct = 0 ; array_ct < std_cone->peak_size; array_ct++) {
		if (std_cone->peaks[array_ct].redundance == HNV_REDUNDANT)
			continue;
		if (std_cone->peaks[array_ct].sum_mod3 != mod_value)
			continue;
		sub_cone->peaks[p_ct].redundance = std_cone->peaks[array_ct].redundance;
		sub_cone->peaks[p_ct].coor.k = std_cone->peaks[array_ct].coor.k;
		sub_cone->peaks[p_ct].coor.l = std_cone->peaks[array_ct].coor.l;
		sub_cone->peaks[p_ct].coor.m = std_cone->peaks[array_ct].coor.m;
		sub_cone->peaks[p_ct].coor.n = std_cone->peaks[array_ct].coor.n;
		sub_cone->peaks[p_ct].sum_mod3 = std_cone->peaks[array_ct].sum_mod3;
		p_ct++;
	}
	sub_cone->peak_ct = p_ct;
	return sub_cone;
}


struct hnv_cone * merge_three_cones(struct hnv_cone * work_cone0, struct hnv_cone * work_cone1,
				    struct hnv_cone * work_cone2)
{
	struct hnv_cone * std_cone;
	int p_size;
	int p_ct;
	int array_ct;
	
	p_size = work_cone0->peak_ct + work_cone1->peak_ct + work_cone2->peak_ct;
	std_cone = allocate_hnv_cone(work_cone0->dim_of_lattice, p_size);
	if (std_cone == NULL) {
		fprintf(STDERR, "merge_three_cones() : ERROR allocation failed (std_cone[%d])\n",
			p_size);
		exit(1);
	}
	
	// Copy peaks
	p_ct = 0;
	for (array_ct = 0 ; array_ct < work_cone0->peak_size; array_ct++) {
		if (work_cone0->peaks[array_ct].redundance == HNV_REDUNDANT)
			continue;
		std_cone->peaks[p_ct].redundance = work_cone0->peaks[array_ct].redundance;
		std_cone->peaks[p_ct].coor.k = work_cone0->peaks[array_ct].coor.k;
		std_cone->peaks[p_ct].coor.l = work_cone0->peaks[array_ct].coor.l;
		std_cone->peaks[p_ct].coor.m = work_cone0->peaks[array_ct].coor.m;
		std_cone->peaks[p_ct].coor.n = work_cone0->peaks[array_ct].coor.n;
		std_cone->peaks[p_ct].sum_mod3 = work_cone0->peaks[array_ct].sum_mod3;
		p_ct++;
	}
	for (array_ct = 0 ; array_ct < work_cone1->peak_size; array_ct++) {
		if (work_cone1->peaks[array_ct].redundance == HNV_REDUNDANT)
			continue;
		std_cone->peaks[p_ct].redundance = work_cone1->peaks[array_ct].redundance;
		std_cone->peaks[p_ct].coor.k = work_cone1->peaks[array_ct].coor.k;
		std_cone->peaks[p_ct].coor.l = work_cone1->peaks[array_ct].coor.l;
		std_cone->peaks[p_ct].coor.m = work_cone1->peaks[array_ct].coor.m;
		std_cone->peaks[p_ct].coor.n = work_cone1->peaks[array_ct].coor.n;
		std_cone->peaks[p_ct].sum_mod3 = work_cone1->peaks[array_ct].sum_mod3;
		p_ct++;
	}
	for (array_ct = 0 ; array_ct < work_cone2->peak_size; array_ct++) {
		if (work_cone2->peaks[array_ct].redundance == HNV_REDUNDANT)
			continue;
		std_cone->peaks[p_ct].redundance = work_cone2->peaks[array_ct].redundance;
		std_cone->peaks[p_ct].coor.k = work_cone2->peaks[array_ct].coor.k;
		std_cone->peaks[p_ct].coor.l = work_cone2->peaks[array_ct].coor.l;
		std_cone->peaks[p_ct].coor.m = work_cone2->peaks[array_ct].coor.m;
		std_cone->peaks[p_ct].coor.n = work_cone2->peaks[array_ct].coor.n;
		std_cone->peaks[p_ct].sum_mod3 = work_cone2->peaks[array_ct].sum_mod3;
		p_ct++;
	}
	std_cone->peak_ct = p_ct;
	return std_cone;
}


//===============================================
// Public methods (part 2)
//===============================================

void copy_hnv_cone(struct hnv_cone * tgt_cone, struct hnv_cone * src_cone)
{
	int p_ct;
	
	if (tgt_cone->peak_size < src_cone->peak_size) {
		fprintf(STDERR, 
			"copy_hnv_cone() : ERROR tgt_cone->peak_size=%d < src_cone->peak_size=%d\n",
			tgt_cone->peak_size, src_cone->peak_size);
		exit(1);
	}
	tgt_cone->dim_of_lattice = src_cone->dim_of_lattice;
	tgt_cone->peak_ct = src_cone->peak_ct;
	for (p_ct = 0; p_ct < src_cone->peak_size; p_ct++) {
		tgt_cone->peaks[p_ct].redundance = src_cone->peaks[p_ct].redundance;
		tgt_cone->peaks[p_ct].coor.k = src_cone->peaks[p_ct].coor.k;
		tgt_cone->peaks[p_ct].coor.l = src_cone->peaks[p_ct].coor.l;
		tgt_cone->peaks[p_ct].coor.m = src_cone->peaks[p_ct].coor.m;
		tgt_cone->peaks[p_ct].coor.n = src_cone->peaks[p_ct].coor.n;
		tgt_cone->peaks[p_ct].sum_mod3 = src_cone->peaks[p_ct].sum_mod3;
	}
	return;
}


HNV_PEAK_TYPE get_peak_redundance(struct hnv_cone * std_cone, int p_ct)
{
	return std_cone->peaks[p_ct].redundance;
}

void set_peak_redundance(struct hnv_cone * std_cone, int p_ct, HNV_PEAK_TYPE value)
{
	std_cone->peaks[p_ct].redundance = value;
	return;
}


int get_peak_sum_mod3(struct hnv_cone * std_cone, int p_ct)
{
	return std_cone->peaks[p_ct].sum_mod3;
}


int get_cone_dim(struct hnv_cone * std_cone)
{
	return std_cone->dim_of_lattice;
}


void set_peak_sum_mod3(struct hnv_cone * std_cone, int p_ct, int value)
{
	int sum_mod;
	
	if (value < 0) 
		sum_mod = value + (get_cone_dim(std_cone) - 1);
	else 
		sum_mod = value;
	std_cone->peaks[p_ct].sum_mod3 = sum_mod;
	return;
}


void print_hnv_cone(struct hnv_cone * std_cone)
{
	int p_ct;
	
	fprintf(STDERR, "print_hnv_cone() : START of cone\n");
	fprintf(STDERR, "print_hnv_cone() : peak_ct = %d, peak_size = %d, dim_of_lattice = %d\n",
		std_cone->peak_ct, std_cone->peak_size, std_cone->dim_of_lattice);
	if (std_cone->peaks != NULL) {
		for (p_ct = 0; p_ct < std_cone->peak_size; p_ct++) {
			// print_hnv_peak(std_cone->peaks + p_ct);
			fprintf(STDERR, "print_hnv_cone() : coor[%d] = { %8.3f, %8.3f, %8.3f, %8.3f }, ", 
				p_ct, std_cone->peaks[p_ct].coor.k, std_cone->peaks[p_ct].coor.l,
				std_cone->peaks[p_ct].coor.m, std_cone->peaks[p_ct].coor.n);
			if (std_cone->peaks[p_ct].sum_mod3 >= 0)
				fprintf(STDERR, " sum mod = %d, ", std_cone->peaks[p_ct].sum_mod3);
			else
				fprintf(STDERR, " sum mod = -, ");
			switch (std_cone->peaks[p_ct].redundance) {
				case HNV_REDUNDANT:
					fprintf(STDOUT, "redundance = REDUNDANT\n");
					break;
				case HNV_NOT_REDUNDANT:
					fprintf(STDOUT, "redundance = NOT_REDUNDANT\n");
					break;
				case HNV_UNKNOWN_P:
					fprintf(STDOUT, "redundance = UNKNOWN\n");
					break;
				case HNV_INVALID_P:
					fprintf(STDOUT, "redundance = INVALID\n");
					break;
				default:
					fprintf(STDOUT, "ERROR redundance = invalid value\n");
					break;
			}
		}
	}
	fprintf(STDERR, "print_hnv_cone() : END of cone\n");
	return;
}


int get_cone_peak_ct(struct hnv_cone * std_cone)
{
	return std_cone->peak_ct;
}


void set_cone_peak_ct(struct hnv_cone * std_cone, int peak_ct)
{
	std_cone->peak_ct = peak_ct;
	return;
}


int get_cone_peak_size(struct hnv_cone * std_cone)
{
	return std_cone->peak_size;
}


void get_peak_coor(struct hnv_coor_4d * peak_coor, struct hnv_cone * std_cone, int p_ct)
{
	peak_coor->k = std_cone->peaks[p_ct].coor.k;
	peak_coor->l = std_cone->peaks[p_ct].coor.l;
	peak_coor->m = std_cone->peaks[p_ct].coor.m;
	peak_coor->n = std_cone->peaks[p_ct].coor.n;
	return;
}


void set_peak_coor(struct hnv_cone * std_cone, int p_ct, struct hnv_coor_4d * peak_coor)
{
	std_cone->peaks[p_ct].coor.k = peak_coor->k;
	std_cone->peaks[p_ct].coor.l = peak_coor->l;
	std_cone->peaks[p_ct].coor.m = peak_coor->m;
	std_cone->peaks[p_ct].coor.n = peak_coor->n;
	return;
}


void load_peaks(struct hnv_cone * standard_cone, FILE * fp_file)
{
	char seq_line[MAX_LINE_SIZE];
	struct hnv_peak * peak_array;
	int peak_ct;
	int seq_ind;
	int coor_sum, coor_sum_mod3;
	
	peak_ct = 0;
	fseek(fp_file, 0L, SEEK_SET);
	while (fgets(seq_line, MAX_LINE_SIZE, fp_file) != NULL) {
		seq_ind = 0;
		while (seq_line[seq_ind] == ' ') seq_ind++;
		if (seq_line[seq_ind] == '{') {	//  seq_line[seq_ind] = "{ x0, y0, z0, w0 }" 
			peak_array = get_cone_peaks(standard_cone);
			load_peak_coordinate(peak_array + peak_ct, seq_line + seq_ind);
			
			coor_sum = (int) (peak_array[peak_ct].coor.k + peak_array[peak_ct].coor.l 
					  + peak_array[peak_ct].coor.m + peak_array[peak_ct].coor.n);
			coor_sum_mod3 = coor_sum % (get_cone_dim(standard_cone) - 1);
			set_peak_sum_mod3(standard_cone, peak_ct, coor_sum_mod3);
			
			set_peak_redundance(standard_cone, peak_ct, HNV_UNKNOWN_P);
			peak_ct++;
		} else if ((seq_line[seq_ind] == '/') && (seq_line[seq_ind + 1] == '/')) {
			continue;	// comment line
		} else {
			continue;
		}
	}
	if (peak_ct != get_cone_peak_size(standard_cone)) {
		fprintf(STDERR, "load_peaks() : ERROR peak_ct=%d  <=> get_cone_peak_ct(standard_cone)=%d\n",
			peak_ct, get_cone_peak_ct(standard_cone));
		exit(1);
	}
	set_cone_peak_ct(standard_cone, peak_ct);
	return;
}


// peak { k, l, m, n } ==>  { k, l, m, n },
//			    { k+1, l, m, n }, { k, l+1, m, n }, { k, l, m+1, n }, { k, l, m, n+1 }, 
//			    { k+2, l, m, n }, { k+1, l+1, m, n }, { k+1, l, m+1, n }, { k+1, l, m, n+1 }, 
//			     ... 
struct hnv_cone * compute_expanded_cone(struct hnv_cone * std_cone)
{
	struct hnv_cone * exp_cone;	// cone of the standard lattice
	float incre[4];
	int p_ct;
	int array_ct;
	int i, j;
	int sum_mod3;
	int mod_value;
	
	exp_cone = (struct hnv_cone *) malloc(sizeof(struct hnv_cone));
	if (exp_cone == NULL) {
		fprintf(STDERR, 
			"compute_expanded_cone() : ERROR allocation failed (exp_cone)\n");
		exit(1);
	}
	initialize_hnv_cone(exp_cone);
	exp_cone->dim_of_lattice = std_cone->dim_of_lattice;
	exp_cone->peak_ct = std_cone->peak_ct + std_cone->peak_ct * 4 + std_cone->peak_ct * 4 * 4;
	exp_cone->peak_size = std_cone->peak_ct + std_cone->peak_ct * 4 + std_cone->peak_ct * 4 * 4;
	exp_cone->peaks = (struct hnv_peak *) malloc(exp_cone->peak_size * sizeof(struct hnv_peak));
	if (exp_cone->peaks == NULL) {
		fprintf(STDERR, "compute_expanded_cone() : ERROR allocation failed (exp_cone->peaks[%d])\n",
			exp_cone->peak_ct);
		exit(1);
	}
	
	for (i = 0; i < 4; i++)
		incre[i] = 0.0;
	p_ct = 0;
	mod_value = std_cone->dim_of_lattice - 1;
	// new peaks = original peak + incre[]
	for (array_ct = 0 ; array_ct < std_cone->peak_size; array_ct++) {
		if (std_cone->peaks[array_ct].redundance == HNV_REDUNDANT)
			continue;
		// { k, l, m, n }
		exp_cone->peaks[p_ct].coor.k = std_cone->peaks[array_ct].coor.k;
		exp_cone->peaks[p_ct].coor.l = std_cone->peaks[array_ct].coor.l;
		exp_cone->peaks[p_ct].coor.m = std_cone->peaks[array_ct].coor.m;
		exp_cone->peaks[p_ct].coor.n = std_cone->peaks[array_ct].coor.n;
		exp_cone->peaks[p_ct].redundance = HNV_UNKNOWN_P;
		exp_cone->peaks[p_ct].sum_mod3 = std_cone->peaks[array_ct].sum_mod3;
		p_ct++;
		for (i = 0; i < 4; i++) {
			// { k, l, m, n } + { ..., 1, ... }
			incre[i]++;
			exp_cone->peaks[p_ct].coor.k = std_cone->peaks[array_ct].coor.k + incre[0];
			exp_cone->peaks[p_ct].coor.l = std_cone->peaks[array_ct].coor.l + incre[1];
			exp_cone->peaks[p_ct].coor.m = std_cone->peaks[array_ct].coor.m + incre[2];
			exp_cone->peaks[p_ct].coor.n = std_cone->peaks[array_ct].coor.n + incre[3];
			exp_cone->peaks[p_ct].redundance = HNV_UNKNOWN_P;
			sum_mod3 = std_cone->peaks[array_ct].sum_mod3 + 1;
			exp_cone->peaks[p_ct].sum_mod3 = sum_mod3 % mod_value;
			p_ct++;
			for (j = 0; j < 4; j++) {
				// { k, l, m, n } + { ..., 1, ..., 1, ... }
				incre[j]++;
				exp_cone->peaks[p_ct].coor.k = std_cone->peaks[array_ct].coor.k + incre[0];
				exp_cone->peaks[p_ct].coor.l = std_cone->peaks[array_ct].coor.l + incre[1];
				exp_cone->peaks[p_ct].coor.m = std_cone->peaks[array_ct].coor.m + incre[2];
				exp_cone->peaks[p_ct].coor.n = std_cone->peaks[array_ct].coor.n + incre[3];
				exp_cone->peaks[p_ct].redundance = HNV_UNKNOWN_P;
				sum_mod3 = std_cone->peaks[array_ct].sum_mod3 + 2;
				exp_cone->peaks[p_ct].sum_mod3 = sum_mod3 % mod_value;
				p_ct++;
				incre[j]--;
			}
			incre[i]--;
		}
	}
	if(p_ct != exp_cone->peak_ct) {
		fprintf(STDERR, 
			"compute_expanded_cone() : ERROR p_ct(=%d) != exp_cone->peak_ct(=%d)\n",
			p_ct, exp_cone->peak_ct);
		exit(1);
	} else {
		if (0) fprintf(STDOUT, 
			"compute_expanded_cone() : DEBUG p_ct(=%d) == exp_cone->peak_ct(=%d)\n",
			p_ct, exp_cone->peak_ct);
	}
	return exp_cone;
}

//
//  x: {1,0,0,0}                  <=>   cx: {0,1,1,1}
//  y: {0,1,0,0}                  <=>   cy: {1,0,1,1}
//  z: {0,0,1,0}                  <=>   cz: {1,1,0,1}
//  w: {0,0,0,1}                  <=>   cw: {1,1,1,0}
//  standard lattice              <=>   conjugate lattice
//  {L+M+N, K+M+N, K+L+N, K+L+M}  <=    {K, L, M, N}
//  {k,l,m,n}                      =>   {(-2k+l+m+n)/3,(k-2l+m+n)/3,  (k+l-2m+n)/3, (k+l+m-2n)/3}
//
struct hnv_cone * compute_conjugate_cone(struct hnv_cone * std_cone)
{
	struct hnv_cone * conj_cone;	// cone of the conjugate lattice
	int p_ct;
	int array_ct;
	
	conj_cone = (struct hnv_cone *) malloc(sizeof(struct hnv_cone));
	if (conj_cone == NULL) {
		fprintf(STDERR, 
			"compute_conjugate_cone() : ERROR allocation failed (conj_cone)\n");
		exit(1);
	}
	initialize_hnv_cone(conj_cone);
	conj_cone->dim_of_lattice = std_cone->dim_of_lattice;
	conj_cone->peak_ct = std_cone->peak_ct;
	conj_cone->peak_size = std_cone->peak_ct;
	conj_cone->peaks = (struct hnv_peak *) malloc(std_cone->peak_ct * sizeof(struct hnv_peak));
	if (conj_cone->peaks == NULL) {
		fprintf(STDERR, 
			"compute_conjugate_cone() : ERROR allocation failed (conj_cone->peaks[%d])\n",
			conj_cone->peak_ct);
		exit(1);
	}
	
	p_ct = 0;
	for (array_ct = 0 ; array_ct < std_cone->peak_size; array_ct++) {
		if (std_cone->peaks[array_ct].redundance == HNV_REDUNDANT)
			continue;
		conj_cone->peaks[p_ct].coor.k
			= (- 2.0*std_cone->peaks[array_ct].coor.k + std_cone->peaks[array_ct].coor.l 
			   + std_cone->peaks[array_ct].coor.m + std_cone->peaks[array_ct].coor.n) / 3.0;
		conj_cone->peaks[p_ct].coor.l
			= (std_cone->peaks[array_ct].coor.k - 2.0*std_cone->peaks[array_ct].coor.l 
			   + std_cone->peaks[array_ct].coor.m + std_cone->peaks[array_ct].coor.n) / 3.0;
		conj_cone->peaks[p_ct].coor.m 
			= (std_cone->peaks[array_ct].coor.k + std_cone->peaks[array_ct].coor.l 
			   - 2.0*std_cone->peaks[array_ct].coor.m + std_cone->peaks[array_ct].coor.n) / 3.0;
		conj_cone->peaks[p_ct].coor.n 
			= (std_cone->peaks[array_ct].coor.k + std_cone->peaks[array_ct].coor.l 
			   + std_cone->peaks[array_ct].coor.m - 2.0*std_cone->peaks[array_ct].coor.n) / 3.0;
		conj_cone->peaks[p_ct].redundance = std_cone->peaks[array_ct].redundance;
		conj_cone->peaks[p_ct].sum_mod3 = std_cone->peaks[array_ct].sum_mod3;
		p_ct++;
	}
	if(p_ct != conj_cone->peak_ct) {
		fprintf(STDERR, 
			"compute_conjugate_cone() : ERROR p_ct(=%d) != conj_cone->peak_ct(=%d)\n",
			p_ct, conj_cone->peak_ct);
		exit(1);
	} else {
		if (0) fprintf(STDOUT, 
			"compute_conjugate_cone() : DEBUG p_ct(=%d) == conj_cone->peak_ct(=%d)\n",
			p_ct, conj_cone->peak_ct);
	}
	return conj_cone;
}

//                    standard    <=    conjugate
//  {L+M+N, K+M+N, K+L+N, K+L+M}  <=    {K, L, M, N}
struct hnv_cone * compute_standard_cone(struct hnv_cone * conj_cone)
{
	struct hnv_cone * std_cone;	// cone of the standard lattice
	int p_ct;
	int array_ct;
	
	std_cone = (struct hnv_cone *) malloc(sizeof(struct hnv_cone));
	if (std_cone == NULL) {
		fprintf(STDERR, 
			"compute_standard_cone() : ERROR allocation failed (std_cone)\n");
		exit(1);
	}
	initialize_hnv_cone(std_cone);
	std_cone->dim_of_lattice = conj_cone->dim_of_lattice;
	std_cone->peak_ct = conj_cone->peak_ct;
	std_cone->peak_size = conj_cone->peak_ct;
	std_cone->peaks = (struct hnv_peak *) malloc(conj_cone->peak_ct * sizeof(struct hnv_peak));
	if (std_cone->peaks == NULL) {
		fprintf(STDERR, "compute_standard_cone() : ERROR allocation failed (std_cone->peaks[%d])\n",
			std_cone->peak_ct);
		exit(1);
	}
	
	p_ct = 0;
	for (array_ct = 0 ; array_ct < conj_cone->peak_size; array_ct++) {
		if (conj_cone->peaks[array_ct].redundance == HNV_REDUNDANT)
			continue;
		std_cone->peaks[p_ct].coor.k 
			= conj_cone->peaks[array_ct].coor.l + conj_cone->peaks[array_ct].coor.m 
			+ conj_cone->peaks[array_ct].coor.n;
		std_cone->peaks[p_ct].coor.l
			= conj_cone->peaks[array_ct].coor.k + conj_cone->peaks[array_ct].coor.m 
			+ conj_cone->peaks[array_ct].coor.n;
		std_cone->peaks[p_ct].coor.m 
			= conj_cone->peaks[array_ct].coor.k + conj_cone->peaks[array_ct].coor.l 
			+ conj_cone->peaks[array_ct].coor.n;
		std_cone->peaks[p_ct].coor.n 
			= conj_cone->peaks[array_ct].coor.k + conj_cone->peaks[array_ct].coor.l 
			+ conj_cone->peaks[array_ct].coor.m;
		std_cone->peaks[p_ct].redundance = conj_cone->peaks[array_ct].redundance;
		std_cone->peaks[p_ct].sum_mod3 = conj_cone->peaks[array_ct].sum_mod3;
		p_ct++;
	}
	if(p_ct != std_cone->peak_ct) {
		fprintf(STDERR, 
			"compute_standard_cone() : ERROR p_ct(=%d) != std_cone->peak_ct(=%d)\n",
			p_ct, std_cone->peak_ct);
		exit(1);
	} else {
		if (0) fprintf(STDOUT, 
			"compute_standard_cone() : DEBUG p_ct(=%d) == std_cone->peak_ct(=%d)\n",
			p_ct, std_cone->peak_ct);
	}
	return std_cone;
}


//
// Height of tangent cone TC[{k, l, m, n}] at {a, b, c}
// {k, l, m, n}+{s, t, u, v} = {(a+b+c)/4, (-a+b-c)/4, (-a-b+c)/4, (a-b-c)/4}+{N, N, N, N}, 
// where s,t,u,v >= 0
//  ==> N = max{ k-( a+b+c)/4, l-(-a+b-c)/4, m-(-a-b+c)/4, n-(a-b-c)/4}
//
float height_std_peak(struct hnv_coor_4d * peak_point, struct hnv_coor_3d * pos)
{
	float Nk, Nl, Nm, Nn;
	float max_Nkl, max_Nmn;
	
	Nk = peak_point->k - (  pos->a + pos->b + pos->c) / 4.0;
	Nl = peak_point->l - (- pos->a + pos->b - pos->c) / 4.0;
	Nm = peak_point->m - (- pos->a - pos->b + pos->c) / 4.0;
	Nn = peak_point->n - (  pos->a - pos->b - pos->c) / 4.0;
	
	if (Nk < Nl)
		max_Nkl = Nl;
	else
		max_Nkl = Nk;
	if (Nm < Nn)
		max_Nmn = Nn;
	else
		max_Nmn = Nm;
	if (max_Nkl < max_Nmn)
		return max_Nmn;
	else
		return max_Nkl;
}


//
// N = min_i{ max{ ki-( a+b+c)/4, li-(-a+b-c)/4, mi-(-a-b+c)/4, ni-(a-b-c)/4 } }
//
float height_std_cone(struct hnv_cone * std_cone, struct hnv_coor_3d * position)
{
	int i;
	int min_i;
	float cur_height;
	float cur_min;
	
	if (0) {
		fprintf(STDOUT, "\n");
		fprintf(STDOUT, "height_std_cone() : DEBUG  START of cone\n");
		print_hnv_cone(std_cone);
		fprintf(STDOUT, "height_std_cone() : DEBUG  END of cone\n");
		fprintf(STDOUT, "\n");
		fprintf(STDOUT, "height_std_cone() : DEBUG position = { %8.3f, %8.3f, %8.3f }\n",
			position->a, position->b, position->c);
		fprintf(STDOUT, "\n");
	}
	i = 0;
	while (std_cone->peaks[i].redundance == HNV_REDUNDANT) i++;
	if (i >= std_cone->peak_size) {
		fprintf(STDERR, "height_std_cone() : ERROR empty cone\n");
		print_hnv_cone(std_cone);
		exit(1);
	}
	cur_min = height_std_peak(&(std_cone->peaks[i].coor), position);
	min_i = i;
	for (i++ ; i < std_cone->peak_size; i++) {
		if (std_cone->peaks[i].redundance == HNV_REDUNDANT)
			continue;
		cur_height = height_std_peak(&(std_cone->peaks[i].coor), position);
		if (cur_min > cur_height) {
			cur_min = cur_height;
			min_i = i;
		}
	}
	if (0) {
		fprintf(STDOUT, "height_std_cone() : DEBUG  cur_min=%f (%d-th peak)\n", cur_min, min_i);
	}
	return cur_min;
}


// 2007-10-06 code tuning of remove_redundant_peaks()
void remove_redundant_peaks(struct hnv_cone * std_cone)
{
	int new_peak_ct;
	int i;
	int p_ct;
	float cur_k, cur_l, cur_m, cur_n;
	int check_result;
	
	new_peak_ct = std_cone->peak_ct;
	for (i = 0; i < std_cone->peak_size; i++) {
		if (new_peak_ct <= 1) {
			if (0) {
				fprintf(STDOUT, "remove_redundant_peaks() : CAUTION one peak cone\n");
				fprintf(STDOUT, "\n");
			}
			if (std_cone->peaks[i].redundance == HNV_UNKNOWN_P) {
				std_cone->peaks[i].redundance = HNV_NOT_REDUNDANT;
				std_cone->peak_ct = new_peak_ct;
				return;
			}
			continue;
		}
		if (std_cone->peaks[i].redundance == HNV_REDUNDANT)
			continue;
		
		// Set current peak
		cur_k = std_cone->peaks[i].coor.k;
		cur_l = std_cone->peaks[i].coor.l;
		cur_m = std_cone->peaks[i].coor.m;
		cur_n = std_cone->peaks[i].coor.n;
		
		std_cone->peaks[i].redundance = HNV_REDUNDANT;	// remove temporary
		check_result = FAIL;
		for (p_ct = 0 ; p_ct < std_cone->peak_size; p_ct++) {
			if (std_cone->peaks[p_ct].redundance == HNV_REDUNDANT)
				continue;
			if ( FLOAT_GE(cur_k, std_cone->peaks[p_ct].coor.k)
			     && FLOAT_GE(cur_l, std_cone->peaks[p_ct].coor.l)
			     && FLOAT_GE(cur_m, std_cone->peaks[p_ct].coor.m)
			     && FLOAT_GE(cur_n, std_cone->peaks[p_ct].coor.n)) {
				check_result = SUCCESS;
				break;	// the cur peak is redundant
			}
		}
		if (check_result == SUCCESS) {
			new_peak_ct--;	// remove i-th peak
		} else {
			if (0) fprintf(STDOUT, "remove_redundant_peaks() : DEBUG %d-th peak removed  (new_peak_ct=%d, peak_size=%d)\n",
				       i, new_peak_ct, std_cone->peak_size);
			std_cone->peaks[i].redundance = HNV_NOT_REDUNDANT;
		}
	}
	std_cone->peak_ct = new_peak_ct;
	
	if (0) {
		fprintf(STDOUT, "remove_redundant_peaks() : DEBUG  START of std_cone\n");
		print_hnv_cone_short(std_cone);
		fprintf(STDOUT, "remove_redundant_peaks() : DEBUG  END of std_cone\n");
		fprintf(STDOUT, "\n");
	}
	if (0) {
		int p_ct;
		int array_ct;
		
		p_ct = 0;
		for (array_ct = 0 ; array_ct < std_cone->peak_size; array_ct++) {
			if (std_cone->peaks[array_ct].redundance == HNV_REDUNDANT)
				continue;
			p_ct++;
		}
		if(p_ct != std_cone->peak_ct) {
			fprintf(STDERR, 
				"remove_redundant_peaks() : ERROR p_ct(=%d) != std_cone->peak_ct(=%d)\n",
				p_ct, std_cone->peak_ct);
			exit(1);
		} else {
			fprintf(STDOUT, 
				"remove_redundant_peaks() : DEBUG p_ct(=%d) == std_cone->peak_ct(=%d)\n",
				p_ct, std_cone->peak_ct);
		}
	}
	return;
}


/* rewrited 2007-10-06 (See above)
void remove_redundant_peaks_old(struct hnv_cone * std_cone)	
{
	int new_peak_ct;
	int i;
	struct hnv_coor_3d cur_position;
	float norm;
	float height;
	
	new_peak_ct = std_cone->peak_ct;
	for (i = 0; i < std_cone->peak_size; i++) {
		if (new_peak_ct <= 1) {	// initial check: if one peak cone or not
			if (0) {
				fprintf(STDOUT, "remove_redundant_peaks() : CAUTION one peak cone\n");
				fprintf(STDOUT, "\n");
			}
			if (std_cone->peaks[i].redundance == HNV_UNKNOWN_P) {
				std_cone->peaks[i].redundance = HNV_NOT_REDUNDANT;
				break;
			}
			continue;
		}
		if (std_cone->peaks[i].redundance == HNV_REDUNDANT)
			continue;
		
		// position = {k-l-m+n, k+l-m-n, k-l+m-n}
		cur_position.a = std_cone->peaks[i].coor.k - std_cone->peaks[i].coor.l 
			- std_cone->peaks[i].coor.m + std_cone->peaks[i].coor.n;
		cur_position.b = std_cone->peaks[i].coor.k + std_cone->peaks[i].coor.l 
			- std_cone->peaks[i].coor.m - std_cone->peaks[i].coor.n;
		cur_position.c = std_cone->peaks[i].coor.k - std_cone->peaks[i].coor.l 
			+ std_cone->peaks[i].coor.m - std_cone->peaks[i].coor.n;
		
		std_cone->peaks[i].redundance = HNV_REDUNDANT;	// remove temporary
		height = height_std_cone(std_cone, &cur_position);
		
		// norm = k+l+m+n
		norm = std_cone->peaks[i].coor.k + std_cone->peaks[i].coor.l 
			+ std_cone->peaks[i].coor.m + std_cone->peaks[i].coor.n;

		if (FLOAT_GT(4.0 * height, norm)) {
			if (0) fprintf(STDOUT, "remove_redundant_peaks() : DEBUG %d-th peak  4.0 * height(=%f) > norm(=%f)\n",
					   i, 4.0 * height, norm);
			std_cone->peaks[i].redundance = HNV_NOT_REDUNDANT;
		} else if ((4.0 * height != norm) && FLOAT_ZERO(4.0 * height - norm)) {
			if (0) fprintf(STDOUT, "remove_redundant_peaks() : DEBUG %d-th peak  4.0 * height(=%f) <=> norm(=%f)\n",
				       i, 4.0 * height, norm);
			new_peak_ct--;	// remove i-th peak
		} else {
			if (0) fprintf(STDOUT, "remove_redundant_peaks() : DEBUG %d-th peak removed  (new_peak_ct=%d, peak_size=%d)\n",
				       i, new_peak_ct, std_cone->peak_size);
			new_peak_ct--;	// remove i-th peak
		}
	}
	std_cone->peak_ct = new_peak_ct;
	
	if (0) {
		fprintf(STDOUT, "remove_redundant_peaks() : DEBUG  START of std_cone\n");
		print_hnv_cone_short(std_cone);
		fprintf(STDOUT, "remove_redundant_peaks() : DEBUG  END of std_cone\n");
		fprintf(STDOUT, "\n");
	}
	if (0) {
		int p_ct;
		int array_ct;
		
		p_ct = 0;
		for (array_ct = 0 ; array_ct < std_cone->peak_size; array_ct++) {
			if (std_cone->peaks[array_ct].redundance == HNV_REDUNDANT)
				continue;
			p_ct++;
		}
		if(p_ct != std_cone->peak_ct) {
			fprintf(STDERR, 
				"remove_redundant_peaks() : ERROR p_ct(=%d) != std_cone->peak_ct(=%d)\n",
				p_ct, std_cone->peak_ct);
			exit(1);
		} else {
			fprintf(STDOUT, 
				"remove_redundant_peaks() : DEBUG p_ct(=%d) == std_cone->peak_ct(=%d)\n",
				p_ct, std_cone->peak_ct);
		}
	}
	return;
}
*/


struct hnv_cone * compute_roof_sub(struct hnv_cone * std_cone)		// Private method
{
	struct hnv_cone * std_roof = NULL;
	
	if (std_cone->peak_ct < 4) {		// roof == cone
		std_roof = allocate_hnv_cone(std_cone->dim_of_lattice, std_cone->peak_ct);
		if (std_roof == NULL) {
			fprintf(STDERR, "compute_roof_sub() : ERROR allocation failed (std_roof[%d])\n",
				std_cone->peak_ct);
			exit(1);
		}
		copy_hnv_cone_compressed(std_roof, std_cone);
	} else {
		int cur_peak_ct;
		int max_peak_ct;
		
		cur_peak_ct = std_cone->peak_ct;
		// max_peak_ct = cur_peak_ct + cur_peak_ct! / ((cur_peak_ct-4)!4!)
		max_peak_ct = cur_peak_ct 
			+ cur_peak_ct * (cur_peak_ct - 1) * (cur_peak_ct - 2) * (cur_peak_ct - 3) / 24;
		std_roof = allocate_hnv_cone(std_cone->dim_of_lattice, max_peak_ct);
		if (std_roof == NULL) {
			fprintf(STDERR, "compute_roof_sub() : ERROR allocation failed (std_roof[%d])\n",
				max_peak_ct);
			exit(1);
		}
		
		compute_roof_for_every_four_peaks(std_roof, std_cone);
		if (DEBUG) {
			// fprintf(STDOUT, "compute_roof_sub() : DEBUG  compute_roof_for_every_four_peaks() DONE (cur_peak_ct=%d, max_peak_ct=%d)\n", cur_peak_ct, max_peak_ct);
			fprintf(STDOUT, "compute_roof_sub() : Now removing redundant peaks of a roof (peak count=%d, peak array size=%d)\n",
				std_roof->peak_ct, std_roof->peak_size);
		}
		remove_redundant_peaks(std_roof);
	}
	if (0) {
		fprintf(STDOUT, "compute_roof_sub() : DEBUG  START of roof\n");
		print_hnv_cone_short(std_roof);
		fprintf(STDOUT, "compute_roof_sub() : DEBUG  END of roof\n");
		fprintf(STDOUT, "\n");
	}
	if (0) {
		int p_ct;
		int array_ct;
		
		p_ct = 0;
		for (array_ct = 0 ; array_ct < std_roof->peak_size; array_ct++) {
			if (std_roof->peaks[array_ct].redundance == HNV_REDUNDANT)
				continue;
			p_ct++;
		}
		if(p_ct != std_roof->peak_ct) {
			fprintf(STDERR, 
				"compute_roof_sub() : ERROR p_ct(=%d) != std_roof->peak_ct(=%d)\n",
				p_ct, std_roof->peak_ct);
			exit(1);
		} else {
			fprintf(STDOUT, 
				"compute_roof_sub() : DEBUG p_ct(=%d) == std_roof->peak_ct(=%d)\n",
				p_ct, std_roof->peak_ct);
		}
	}
	return std_roof;
}


//
// Tangent roof TR[p1, p2, p3, p4], where pi={ki, li, mi, ni}
// TR[p0, p1, p2, p3] = TC[p] (p= {k, l, m, n}), where
//    k = the third smallest of {ki},
//    l = the third smallest of {li},
//    m = the third smallest of {mi},
//    n = the third smallest of {ni}
//
struct hnv_cone * compute_roof(struct hnv_cone * std_cone)
{
	struct hnv_cone * work_cones[3];
	struct hnv_cone * work_roofs[3];
	struct hnv_cone * std_roof = NULL;
	int mod_value;
	
	if (std_cone->peak_ct < 4) {		// roof == cone
		std_roof = allocate_hnv_cone(std_cone->dim_of_lattice, std_cone->peak_ct);
		if (std_roof == NULL) {
			fprintf(STDERR, "compute_roof() : ERROR allocation failed (std_roof[%d])\n",
				std_cone->peak_ct);
			exit(1);
		}
		copy_hnv_cone_compressed(std_roof, std_cone);
		remove_redundant_peaks(std_roof);
	} else {
		if (get_use_extended_conj_cone() == HNV_TRUE) {
			// Compute roof for each sum_mod3
			for (mod_value = 0; mod_value < 3; mod_value++) {
				work_cones[mod_value] = extract_sub_cone(std_cone, mod_value);
				if (DEBUG)
					fprintf(STDOUT, "compute_roof() : Compute roof (peak count=%d, peak array size=%d, sum=%d mod 3)\n",
						work_cones[mod_value]->peak_ct, work_cones[mod_value]->peak_ct, mod_value);
				work_roofs[mod_value] = compute_roof_sub(work_cones[mod_value]);
			}
			
			{	// peak count check
				int sum_of_peak_ct;
				
				sum_of_peak_ct = work_cones[0]->peak_ct + work_cones[1]->peak_ct + work_cones[2]->peak_ct;
				if (std_cone->peak_ct != sum_of_peak_ct) {
					fprintf(STDERR, "compute_roof() : ERROR peak count check  std_cone->peak_ct(=%d) != sum_of_peak_ct(=%d)\n",
						std_roof->peak_ct, sum_of_peak_ct);
					fprintf(STDERR, "compute_roof() : ERROR work_cones[0]->peak_ct=%d, work_cones[1]->peak_ct=%d, work_cones[2]->peak_ct=%d\n",
						work_cones[0]->peak_ct, work_cones[1]->peak_ct, work_cones[2]->peak_ct);
					exit(1);
				}
			}
			
			std_roof = merge_three_cones(work_roofs[0], work_roofs[1], work_roofs[2]);
			if (std_roof->peak_ct <= 0) {
				fprintf(STDERR, "compute_roof() : ERROR empty std_roof(ct=%d, size=%d)\n",
					std_roof->peak_ct, std_roof->peak_size);
				exit(1);
			}
			remove_redundant_peaks(std_roof);
			
			for (mod_value = 0; mod_value < 3; mod_value++) {
				free_hnv_cone(work_cones[mod_value]);
				free_hnv_cone(work_roofs[mod_value]);
			}
		} else {
			std_roof =compute_roof_sub(std_cone);
		}
	}
	if (0) {
		fprintf(STDOUT, "compute_roof() : DEBUG  START of roof\n");
		print_hnv_cone(std_roof);
		fprintf(STDOUT, "compute_roof() : DEBUG  END of roof\n");
		fprintf(STDOUT, "\n");
	}
	
	if (DEBUG) {
		int p_ct;
		int array_ct;
		
		p_ct = 0;
		for (array_ct = 0 ; array_ct < std_roof->peak_size; array_ct++) {
			if (std_roof->peaks[array_ct].redundance == HNV_REDUNDANT)
				continue;
			p_ct++;
		}
		if(p_ct != std_roof->peak_ct) {
			fprintf(STDERR, 
				"compute_roof() : ERROR p_ct(=%d) != std_roof->peak_ct(=%d), std_cone->peak_size=%d\n",
				p_ct, std_roof->peak_ct, std_cone->peak_size);
			exit(1);
		} else {
			if (0) fprintf(STDOUT, 
				"compute_roof() : DEBUG p_ct(=%d) == std_roof->peak_ct(=%d), std_cone->peak_size=%d\n",
				p_ct, std_roof->peak_ct, std_cone->peak_size);
		}
	}
	return std_roof;
}


//
// peak_point = {k, l, m, n}
// position = {k0, l0, m0, n0}
// {k, l, m, n}+{0,s,s,s}+{t,0,t,t}+{u,u,0,u}+{v,v,v,0} (= {k, l, m, n}+{t+u+v,s+u+v,s+t+v,s+t+u})
// = {(a+b+c)/4, (-a+b-c)/4, (-a-b+c)/4, (a-b-c)/4}+{N,N,N,N}, where s,t,u,v>0
// =>
//   N = max { -2k+l+m+n + 3(a+b+c)/4, k-2l+m+n + 3(-a+b-c)/4,
//             k+l-2m+n + 3(-a-b+c)/4, k+l+m-2n + 3(a-b-c)/4 }
//
float height_conj_peak(struct hnv_coor_4d * peak_point, struct hnv_coor_3d * pos)
{
	float Nk, Nl, Nm, Nn;
	float max_kl, max_mn;
	
	Nk = (- 2.0*peak_point->k + peak_point->l + peak_point->m + peak_point->n)
		+ 3.0 * (  pos->a + pos->b + pos->c) / 4.0;
	Nl = (peak_point->k - 2.0*peak_point->l + peak_point->m + peak_point->n)
		+ 3.0 * (- pos->a + pos->b - pos->c) / 4.0;
	Nm = (peak_point->k + peak_point->l - 2.0* peak_point->m + peak_point->n)
		+ 3.0 * (- pos->a - pos->b + pos->c) / 4.0;
	Nn = (peak_point->k + peak_point->l + peak_point->m  - 2.0*peak_point->n)
		+ 3.0 * (  pos->a - pos->b - pos->c) / 4.0;
	
	if (Nk < Nl)
		max_kl = Nl;
	else
		max_kl = Nk;
	if (Nm < Nn)
		max_mn = Nn;
	else
		max_mn = Nm;
	if (max_kl < max_mn)
		return max_mn;
	else
		return max_kl;
}


//
//  N = min { max{ -2k+l+m+n + 3(a+b+c)/4, k-2l+m+n + 3(-a+b-c)/4,
//             k+l-2m+n + 3(-a-b+c)/4, k+l+m-2n + 3(a-b-c)/4 } }
//
float height_conj_cone(struct hnv_cone * conj_cone, struct hnv_coor_3d * position)
{
	int i;
	int min_i;
	float cur_height;
	float cur_min;
	
	i = 0;
	while (conj_cone->peaks[i].redundance == HNV_REDUNDANT) i++;
	if (i >= conj_cone->peak_size) {
		fprintf(STDERR, "height_conj_cone() : ERROR empty cone\n");
		exit(1);
	}
	cur_min = height_conj_peak(&(conj_cone->peaks[i].coor), position);
	min_i = i;
	for (i++ ; i < conj_cone->peak_size; i++) {
		if (conj_cone->peaks[i].redundance == HNV_REDUNDANT)
			continue;
		cur_height = height_conj_peak(&(conj_cone->peaks[i].coor), position);
		if (cur_min > cur_height) {
			cur_min = cur_height;
			min_i = i;
		}
	}
	if (0) {
		fprintf(STDOUT, "height_conj_cone() : DEBUG  cur_min=%f (%d-th peak)\n", cur_min, min_i);
	}
	return cur_min;
}


struct hnv_cone * compute_conjugate_roof(struct hnv_cone * std_cone)
{
	struct hnv_cone * conj_cone;		// cone of the conjugate lattice
	struct hnv_cone * conj_roof;		// cone of the conjugate lattice
	struct hnv_cone * conjugate_roof;	// cone of the standard lattice
	
	conj_cone = compute_conjugate_cone(std_cone);
	remove_redundant_peaks(conj_cone);
	if (0) {
		fprintf(STDOUT, "compute_conjugate_roof() : DEBUG  START of conj_cone\n");
		print_hnv_cone_short(conj_cone);
		fprintf(STDOUT, "compute_conjugate_roof() : DEBUG  END of conj_cone\n");
		fprintf(STDOUT, "\n");
	}
	conj_roof = compute_roof(conj_cone);
	if (0) {
		fprintf(STDOUT, "compute_conjugate_roof() : DEBUG  START of conj_roof\n");
		print_hnv_cone_short(conj_roof);
		fprintf(STDOUT, "compute_conjugate_roof() : DEBUG  END of conj_roof\n");
		fprintf(STDOUT, "\n");
	}
	conjugate_roof = compute_standard_cone(conj_roof);	// conj_roof in standard lattice
	if (0) {
		fprintf(STDOUT, "compute_conjugate_roof() : DEBUG  START of conjugate_roof\n");
		print_hnv_cone_short(conjugate_roof);
		fprintf(STDOUT, "compute_conjugate_roof() : DEBUG  END of conjugate_roof\n");
		fprintf(STDOUT, "\n");
	}
	free_hnv_cone(conj_cone);
	free_hnv_cone(conj_roof);
	return conjugate_roof;
}

