/*
 *  hnv_load_file.c
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/15/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 */

#include <stdio.h>
#include <stdlib.h>
#include "hnv_main.h"
#include "hnv_computation_main.h"
#include "hnv_cone.h"

//===============================================
// Private data
//===============================================

//===============================================
// Public data
//===============================================

// [Input file: "xxxx.hnv"]
//   //  comments
//   { x0, y0, z0, w0 }    //  position of peaks
//   { x1, y1, z1, w1 }


//===============================================
// Private methods
//===============================================
//
// struct hnv_cone * read_and_allocate_cone(FILE * fp_file);
//-----------------------------------------------

struct hnv_cone * read_and_allocate_cone(FILE * fp_file)
{
	struct hnv_cone * standard_cone;
	char seq_line[MAX_LINE_SIZE];
	int seq_ind = 0;
	int dim_ct;
	int peak_ct;
	
	// Check dimension of lattice
	dim_ct = 0;
	while (fgets(seq_line, MAX_LINE_SIZE, fp_file) != NULL) {
		seq_ind = 0;
		while (seq_line[seq_ind] == ' ') seq_ind++;
		if (seq_line[seq_ind] == '{') {
			seq_ind++;
			break;
		} else if ((seq_line[seq_ind] == '/') && (seq_line[seq_ind + 1] == '/')) {
			continue;	// comment line
		} else {
			continue;
		}
	}
	while (seq_line[seq_ind] != '\n') {
		while (seq_line[seq_ind] == ' ') seq_ind++;
		if (seq_line[seq_ind] == ',') {
			dim_ct++;
		} else if (seq_line[seq_ind] == '}') {
			dim_ct++;
			break;
		}
		seq_ind++;
	}
	if (dim_ct != 4) {
		fprintf(STDERR, "read_and_allocate_cone() : ERROR dim = %d not supported\n", dim_ct);
		fprintf(STDERR, "read_and_allocate_cone() : ERROR seq_line = \"%s\"\n", seq_line);
		return (struct hnv_cone *) NULL;
	}

	// Count peak number
	peak_ct = 0;
	fseek(fp_file, 0L, SEEK_SET);
	while (fgets(seq_line, MAX_LINE_SIZE, fp_file) != NULL) {
		seq_ind = 0;
		while (seq_line[seq_ind] == ' ') seq_ind++;
		if (seq_line[seq_ind] == '{') {
			peak_ct++;
		} else if ((seq_line[seq_ind] == '/') && (seq_line[seq_ind + 1] == '/')) {
			continue;	// comment line
		} else {
			continue;
		}
	}
	standard_cone = allocate_hnv_cone(dim_ct, peak_ct);
	if (standard_cone == NULL) {
		fprintf(STDERR, "read_and_allocate_cone() : ERROR allocation failed (standard_cone[%d])\n",
			peak_ct);
		exit(1);
	}
	
	return standard_cone;
}


//===============================================
// Public methods
//===============================================

struct hnv_cone * load_hnv_file(const char * filename)
{
	FILE * hnv_file;
	struct hnv_cone * standard_cone;
	
	hnv_file = fopen(filename, "r");
	if (hnv_file == NULL) {
		fprintf(STDERR, "load_hnv_file() : ERROR can't open file \"%s\"\n", filename);
		exit(1);
	}
	standard_cone = read_and_allocate_cone(hnv_file);
	if (standard_cone == NULL) {
		fprintf(STDERR, "load_hnv_file() : ERROR can't load standard cone from \"%s\"\n", filename);
		exit(1);
	}
	
	load_peaks(standard_cone, hnv_file);
	if (fclose(hnv_file)) {
		fprintf(STDERR, "load_hnv_file() : ERROR can't close file \"%s\"\n", filename);
		abort();
	}
	return standard_cone;
}

