/*
 *  hnv_main.h
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/9/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 *  2007-10-03 relax the condition of FLOAT_GT and FLOAT_ZERO
 *  2007-10-03 version 0.1a release
 *  2007-10-06 add FLOAT_GE
 */

#define STDOUT stdout
#define STDERR stdout
#define DEBUG 1

#define SUCCESS 0
#define FAIL 1
#define YES 0
#define NONO 1

#define MAX_LINE_SIZE  256

/*
 Window coor (x,y)
 (0,0)   ->    (640,0)
   |
  \/
 (0,480) ->  (640,480)
 */
#define WindowWidth 640
#define WindowHeight 480
#define ON 1
#define OFF -1

#define PI 3.14159265358979323846  // ==> math.h
// 2007-10-03 relax the condition of FLOAT_GT and FLOAT_ZERO
//#define FLOAT_GT(fdata1, fdata2) ((fdata1) > (fdata2) + 1.0 / 1000000.0)
//#define FLOAT_ZERO(fdata) ((fdata) * (fdata) <  1.0 / 1000000.0)
#define FLOAT_GT(fdata1, fdata2) ((fdata1) > (fdata2) + 1.0 / 10000.0)
#define FLOAT_ZERO(fdata) ((fdata) * (fdata) <  1.0 / 10000.0)
// 2007-10-06
#define FLOAT_GE(fdata1, fdata2) ((fdata1) > (fdata2) - 1.0 / 10000.0)

// tetrahedron direction
#define HNV_UP   1
#define HNV_DOWN   (-1)

typedef enum hnv_boolean {
	HNV_TRUE,
	HNV_FALSE
} HNV_BOOLEAN;

typedef enum hnv_gradient {
	HNV_UNKNOWN_G,
	HNV_XYZ,
	HNV_XYW,
	HNV_XZW,
	HNV_YZW
} HNV_GRADIENT;

typedef enum hnv_tetra_class {
	HNV_UNKNOWN_C,
	HNV_IN,
	HNV_OUT,
	HNV_BD
} HNV_TETRA_CLASS;

typedef enum hnv_tetra_chiral {
	HNV_UNKNOWN_CHIRAL,
	HNV_RIGHT_CHIRAL,
	HNV_LEFT_CHIRAL
} HNV_TETRA_CHIRAL;

typedef enum hnv_traj_type {
	HNV_UNKNOWN_T,
	HNV_CLOSED,
	HNV_NOT_CLOSED,
	HNV_ALL
} HNV_TRAJ_TYPE;

typedef enum hnv_visible_peaks {
	HNV_UNKNOWN_PEAKS,
	HNV_STD_CONE,
	HNV_STD_ROOF,
	HNV_CONJ_ROOF
} HNV_VISIBLE_PEAKS;

typedef enum hnv_peak_type {
	HNV_UNKNOWN_P,
	HNV_REDUNDANT,
	HNV_NOT_REDUNDANT,
	HNV_INVALID_P
} HNV_PEAK_TYPE;

typedef enum hnv_compare_result {
	HNV_EQUAL,
	HNV_LEFT,
	HNV_RIGHT
} HNV_COMPARE_RESULT;		// used when snode_tree is searched

struct hnv_coor_3d {
	float a;
	float b;
	float c;
};

struct hnv_coor_3d_int {
	int a;
	int b;
	int c;
};

struct hnv_coor_4d {
	float k;
	float l;
	float m;
	float n;
};