/*
 *  hnv_tetra_around_peak.c
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/16/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 */

#include <stdio.h>
#include "hnv_main.h"
#include "hnv_computation_main.h"
#include "hnv_tetrahedron.h"

struct tetra_vertices {
	struct hnv_coor_3d top;
	struct hnv_coor_3d second;
	struct hnv_coor_3d third;
	struct hnv_coor_3d bottom;
} tetras_around_peak[24];


//===============================================
// Private data
//===============================================

//===============================================
// Private methods
//===============================================
//
//  void tetrahedrons_around_origin(void);
//-----------------------------------------------

//
// tetrahedrons around peak p={0, 0, 0, 0}
//     x: {1,0,0,0} -->==> { 1,  1,  1}
//     y: {0,1,0,0} -->==> {-1,  1, -1}
//     z: {0,0,1,0} -->==> {-1, -1,  1}
//     w: {0,0,0,1} -->==> { 1, -1, -1}
//
void tetrahedrons_around_origin(void)
{
//	p[xyz] -->==> { {0,0,0}, { 1, 1, 1}, { 0, 2, 0}, {-1, 1, 1} }
	tetras_around_peak[0].top.a = 0.0;
	tetras_around_peak[0].top.b = 0.0;
	tetras_around_peak[0].top.c = 0.0;
	tetras_around_peak[0].second.a = 1.0;
	tetras_around_peak[0].second.b = 1.0;
	tetras_around_peak[0].second.c = 1.0;
	tetras_around_peak[0].third.a = 0.0;
	tetras_around_peak[0].third.b = 2.0;
	tetras_around_peak[0].third.c = 0.0;
	tetras_around_peak[0].bottom.a = -1.0;
	tetras_around_peak[0].bottom.b = 1.0;
	tetras_around_peak[0].bottom.c = 1.0;
//	p[xzy] -->==> { {0,0,0}, { 1, 1, 1}, { 0, 0, 2}, {-1, 1, 1} }
	tetras_around_peak[1].top.a = 0.0;
	tetras_around_peak[1].top.b = 0.0;
	tetras_around_peak[1].top.c = 0.0;
	tetras_around_peak[1].second.a = 1.0;
	tetras_around_peak[1].second.b = 1.0;
	tetras_around_peak[1].second.c = 1.0;
	tetras_around_peak[1].third.a = 0.0;
	tetras_around_peak[1].third.b = 0.0;
	tetras_around_peak[1].third.c = 2.0;
	tetras_around_peak[1].bottom.a = -1.0;
	tetras_around_peak[1].bottom.b = 1.0;
	tetras_around_peak[1].bottom.c = 1.0;
//	p[yxz] -->==> { {0,0,0}, {-1, 1,-1}, { 0, 2, 0}, {-1, 1, 1} }
	tetras_around_peak[2].top.a = 0.0;
	tetras_around_peak[2].top.b = 0.0;
	tetras_around_peak[2].top.c = 0.0;
	tetras_around_peak[2].second.a = -1.0;
	tetras_around_peak[2].second.b = 1.0;
	tetras_around_peak[2].second.c = -1.0;
	tetras_around_peak[2].third.a = 0.0;
	tetras_around_peak[2].third.b = 2.0;
	tetras_around_peak[2].third.c = 0.0;
	tetras_around_peak[2].bottom.a = -1.0;
	tetras_around_peak[2].bottom.b = 1.0;
	tetras_around_peak[2].bottom.c = 1.0;
//	p[yzx] -->==> { {0,0,0}, {-1, 1,-1}, {-2, 0, 0}, {-1, 1, 1} }
	tetras_around_peak[3].top.a = 0.0;
	tetras_around_peak[3].top.b = 0.0;
	tetras_around_peak[3].top.c = 0.0;
	tetras_around_peak[3].second.a = -1.0;
	tetras_around_peak[3].second.b = 1.0;
	tetras_around_peak[3].second.c = -1.0;
	tetras_around_peak[3].third.a = -2.0;
	tetras_around_peak[3].third.b = 0.0;
	tetras_around_peak[3].third.c = 0.0;
	tetras_around_peak[3].bottom.a = -1.0;
	tetras_around_peak[3].bottom.b = 1.0;
	tetras_around_peak[3].bottom.c = 1.0;
//	p[zxy] -->==> { {0,,0}, {-1,-1, 1}, { 0, 0, 2}, {-1, 1, 1} }
	tetras_around_peak[4].top.a = 0.0;
	tetras_around_peak[4].top.b = 0.0;
	tetras_around_peak[4].top.c = 0.0;
	tetras_around_peak[4].second.a = -1.0;
	tetras_around_peak[4].second.b = -1.0;
	tetras_around_peak[4].second.c = 1.0;
	tetras_around_peak[4].third.a = 0.0;
	tetras_around_peak[4].third.b = 0.0;
	tetras_around_peak[4].third.c = 2.0;
	tetras_around_peak[4].bottom.a = -1.0;
	tetras_around_peak[4].bottom.b = 1.0;
	tetras_around_peak[4].bottom.c = 1.0;
//	p[zyx] -->==> { {0,0,0}, {-1,-1, 1}, {-2, 0, 0}, {-1, 1, 1} }
	tetras_around_peak[5].top.a = 0.0;
	tetras_around_peak[5].top.b = 0.0;
	tetras_around_peak[5].top.c = 0.0;
	tetras_around_peak[5].second.a = -1.0;
	tetras_around_peak[5].second.b = -1.0;
	tetras_around_peak[5].second.c = 1.0;
	tetras_around_peak[5].third.a = -2.0;
	tetras_around_peak[5].third.b = 0.0;
	tetras_around_peak[5].third.c = 0.0;
	tetras_around_peak[5].bottom.a = -1.0;
	tetras_around_peak[5].bottom.b = 1.0;
	tetras_around_peak[5].bottom.c = 1.0;
//---
//	p[xyw] -->==> { {0,0,0}, { 1, 1, 1}, { 0, 2, 0}, { 1, 1,-1} }
	tetras_around_peak[6].top.a = 0.0;
	tetras_around_peak[6].top.b = 0.0;
	tetras_around_peak[6].top.c = 0.0;
	tetras_around_peak[6].second.a = 1.0;
	tetras_around_peak[6].second.b = 1.0;
	tetras_around_peak[6].second.c = 1.0;
	tetras_around_peak[6].third.a = 0.0;
	tetras_around_peak[6].third.b = 2.0;
	tetras_around_peak[6].third.c = 0.0;
	tetras_around_peak[6].bottom.a = 1.0;
	tetras_around_peak[6].bottom.b = 1.0;
	tetras_around_peak[6].bottom.c = -1.0;
//	p[xwy] -->==> { {0,0,0}, { 1, 1, 1}, { 2, 0, 0}, { 1, 1,-1} }
	tetras_around_peak[7].top.a = 0.0;
	tetras_around_peak[7].top.b = 0.0;
	tetras_around_peak[7].top.c = 0.0;
	tetras_around_peak[7].second.a = 1.0;
	tetras_around_peak[7].second.b = 1.0;
	tetras_around_peak[7].second.c = 1.0;
	tetras_around_peak[7].third.a = 2.0;
	tetras_around_peak[7].third.b = 0.0;
	tetras_around_peak[7].third.c = 0.0;
	tetras_around_peak[7].bottom.a = 1.0;
	tetras_around_peak[7].bottom.b = 1.0;
	tetras_around_peak[7].bottom.c = -1.0;
//	p[yxw] -->==> { {0,0,0}, {-1, 1,-1}, { 0, 2, 0}, { 1, 1,-1} }
	tetras_around_peak[8].top.a = 0.0;
	tetras_around_peak[8].top.b = 0.0;
	tetras_around_peak[8].top.c = 0.0;
	tetras_around_peak[8].second.a = -1.0;
	tetras_around_peak[8].second.b = 1.0;
	tetras_around_peak[8].second.c = -1.0;
	tetras_around_peak[8].third.a = 0.0;
	tetras_around_peak[8].third.b = 2.0;
	tetras_around_peak[8].third.c = 0.0;
	tetras_around_peak[8].bottom.a = 1.0;
	tetras_around_peak[8].bottom.b = 1.0;
	tetras_around_peak[8].bottom.c = -1.0;
//	p[ywx] -->==> { {0,0,0}, {-1, 1,-1}, { 0, 0,-2}, { 1, 1,-1} }
	tetras_around_peak[9].top.a = 0.0;
	tetras_around_peak[9].top.b = 0.0;
	tetras_around_peak[9].top.c = 0.0;
	tetras_around_peak[9].second.a = -1.0;
	tetras_around_peak[9].second.b = 1.0;
	tetras_around_peak[9].second.c = -1.0;
	tetras_around_peak[9].third.a = 0.0;
	tetras_around_peak[9].third.b = 0.0;
	tetras_around_peak[9].third.c = -2.0;
	tetras_around_peak[9].bottom.a = 1.0;
	tetras_around_peak[9].bottom.b = 1.0;
	tetras_around_peak[9].bottom.c = -1.0;
//	p[wxy] -->==> { {0,0,0}, { 1,-1,-1}, { 2, 0, 0}, { 1, 1,-1} }
	tetras_around_peak[10].top.a = 0.0;
	tetras_around_peak[10].top.b = 0.0;
	tetras_around_peak[10].top.c = 0.0;
	tetras_around_peak[10].second.a = 1.0;
	tetras_around_peak[10].second.b = -1.0;
	tetras_around_peak[10].second.c = -1.0;
	tetras_around_peak[10].third.a = 2.0;
	tetras_around_peak[10].third.b = 0.0;
	tetras_around_peak[10].third.c = 0.0;
	tetras_around_peak[10].bottom.a = 1.0;
	tetras_around_peak[10].bottom.b = 1.0;
	tetras_around_peak[10].bottom.c = -1.0;
//	p[wyx] -->==> { {0,0,0}, { 1,-1,-1}, { 0, 0,-2}, { 1, 1,-1} }
	tetras_around_peak[11].top.a = 0.0;
	tetras_around_peak[11].top.b = 0.0;
	tetras_around_peak[11].top.c = 0.0;
	tetras_around_peak[11].second.a = 1.0;
	tetras_around_peak[11].second.b = -1.0;
	tetras_around_peak[11].second.c = -1.0;
	tetras_around_peak[11].third.a = 0.0;
	tetras_around_peak[11].third.b = 0.0;
	tetras_around_peak[11].third.c = -2.0;
	tetras_around_peak[11].bottom.a = 1.0;
	tetras_around_peak[11].bottom.b = 1.0;
	tetras_around_peak[11].bottom.c = -1.0;
//---
//	p[xzw] -->==> { {0,0,0}, { 1, 1, 1}, { 0, 0, 2}, { 1,-1, 1} }
	tetras_around_peak[12].top.a = 0.0;
	tetras_around_peak[12].top.b = 0.0;
	tetras_around_peak[12].top.c = 0.0;
	tetras_around_peak[12].second.a = 1.0;
	tetras_around_peak[12].second.b = 1.0;
	tetras_around_peak[12].second.c = 1.0;
	tetras_around_peak[12].third.a = 0.0;
	tetras_around_peak[12].third.b = 0.0;
	tetras_around_peak[12].third.c = 2.0;
	tetras_around_peak[12].bottom.a = 1.0;
	tetras_around_peak[12].bottom.b = -1.0;
	tetras_around_peak[12].bottom.c = 1.0;
//	p[xwz] -->==> { {0,0,0}, { 1, 1, 1}, { 2, 0, 0}, { 1,-1, 1} }
	tetras_around_peak[13].top.a = 0.0;
	tetras_around_peak[13].top.b = 0.0;
	tetras_around_peak[13].top.c = 0.0;
	tetras_around_peak[13].second.a = 1.0;
	tetras_around_peak[13].second.b = 1.0;
	tetras_around_peak[13].second.c = 1.0;
	tetras_around_peak[13].third.a = 2.0;
	tetras_around_peak[13].third.b = 0.0;
	tetras_around_peak[13].third.c = 0.0;
	tetras_around_peak[13].bottom.a = 1.0;
	tetras_around_peak[13].bottom.b = -1.0;
	tetras_around_peak[13].bottom.c = 1.0;
//	p[zxw] -->==> { {0,0,0}, {-1,-1, 1}, { 0, 0, 2}, { 1,-1, 1} }
	tetras_around_peak[14].top.a = 0.0;
	tetras_around_peak[14].top.b = 0.0;
	tetras_around_peak[14].top.c = 0.0;
	tetras_around_peak[14].second.a = -1.0;
	tetras_around_peak[14].second.b = -1.0;
	tetras_around_peak[14].second.c = 1.0;
	tetras_around_peak[14].third.a = 0.0;
	tetras_around_peak[14].third.b = 0.0;
	tetras_around_peak[14].third.c = 2.0;
	tetras_around_peak[14].bottom.a = 1.0;
	tetras_around_peak[14].bottom.b = -1.0;
	tetras_around_peak[14].bottom.c = 1.0;
//	p[zwx] -->==> { {0,0,0}, {-1,-1, 1}, { 0,-2, 0}, { 1,-1, 1} }
	tetras_around_peak[15].top.a = 0.0;
	tetras_around_peak[15].top.b = 0.0;
	tetras_around_peak[15].top.c = 0.0;
	tetras_around_peak[15].second.a = -1.0;
	tetras_around_peak[15].second.b = -1.0;
	tetras_around_peak[15].second.c = 1.0;
	tetras_around_peak[15].third.a = 0.0;
	tetras_around_peak[15].third.b = -2.0;
	tetras_around_peak[15].third.c = 0.0;
	tetras_around_peak[15].bottom.a = 1.0;
	tetras_around_peak[15].bottom.b = -1.0;
	tetras_around_peak[15].bottom.c = 1.0;
//	p[wxz] -->==> { {0,0,0}, { 1,-1,-1}, { 2, 0, 0}, { 1,-1, 1} }
	tetras_around_peak[16].top.a = 0.0;
	tetras_around_peak[16].top.b = 0.0;
	tetras_around_peak[16].top.c = 0.0;
	tetras_around_peak[16].second.a = 1.0;
	tetras_around_peak[16].second.b = -1.0;
	tetras_around_peak[16].second.c = -1.0;
	tetras_around_peak[16].third.a = 2.0;
	tetras_around_peak[16].third.b = 0.0;
	tetras_around_peak[16].third.c = 0.0;
	tetras_around_peak[16].bottom.a = 1.0;
	tetras_around_peak[16].bottom.b = -1.0;
	tetras_around_peak[16].bottom.c = 1.0;
//	p[wzx] -->==> { {0,0,0}, { 1,-1,-1}, { 0,-2, 0}, { 1,-1, 1} }
	tetras_around_peak[17].top.a = 0.0;
	tetras_around_peak[17].top.b = 0.0;
	tetras_around_peak[17].top.c = 0.0;
	tetras_around_peak[17].second.a = 1.0;
	tetras_around_peak[17].second.b = -1.0;
	tetras_around_peak[17].second.c = -1.0;
	tetras_around_peak[17].third.a = 0.0;
	tetras_around_peak[17].third.b = -2.0;
	tetras_around_peak[17].third.c = 0.0;
	tetras_around_peak[17].bottom.a = 1.0;
	tetras_around_peak[17].bottom.b = -1.0;
	tetras_around_peak[17].bottom.c = 1.0;
//---
//	p[yzw] -->==> { {0,0,0}, {-1, 1,-1}, {-2, 0, 0}, {-1,-1,-1} }
	tetras_around_peak[18].top.a = 0.0;
	tetras_around_peak[18].top.b = 0.0;
	tetras_around_peak[18].top.c = 0.0;
	tetras_around_peak[18].second.a = -1.0;
	tetras_around_peak[18].second.b = 1.0;
	tetras_around_peak[18].second.c = -1.0;
	tetras_around_peak[18].third.a = -2.0;
	tetras_around_peak[18].third.b = 0.0;
	tetras_around_peak[18].third.c = 0.0;
	tetras_around_peak[18].bottom.a = -1.0;
	tetras_around_peak[18].bottom.b = -1.0;
	tetras_around_peak[18].bottom.c = -1.0;
//	p[ywz] -->==> { {0,0,0}, {-1, 1,-1}, { 0, 0,-2}, {-1,-1,-1} }
	tetras_around_peak[19].top.a = 0.0;
	tetras_around_peak[19].top.b = 0.0;
	tetras_around_peak[19].top.c = 0.0;
	tetras_around_peak[19].second.a = -1.0;
	tetras_around_peak[19].second.b = 1.0;
	tetras_around_peak[19].second.c = -1.0;
	tetras_around_peak[19].third.a = 0.0;
	tetras_around_peak[19].third.b = 0.0;
	tetras_around_peak[19].third.c = -2.0;
	tetras_around_peak[19].bottom.a = -1.0;
	tetras_around_peak[19].bottom.b = -1.0;
	tetras_around_peak[19].bottom.c = -1.0;
//	p[zyw] -->==> { {0,0,0}, {-1,-1, 1}, {-2, 0, 0}, {-1,-1,-1} }
	tetras_around_peak[20].top.a = 0.0;
	tetras_around_peak[20].top.b = 0.0;
	tetras_around_peak[20].top.c = 0.0;
	tetras_around_peak[20].second.a = -1.0;
	tetras_around_peak[20].second.b = -1.0;
	tetras_around_peak[20].second.c = 1.0;
	tetras_around_peak[20].third.a = -2.0;
	tetras_around_peak[20].third.b = 0.0;
	tetras_around_peak[20].third.c = 0.0;
	tetras_around_peak[20].bottom.a = -1.0;
	tetras_around_peak[20].bottom.b = -1.0;
	tetras_around_peak[20].bottom.c = -1.0;
//	p[zwy] -->==> { {0,0,0}, {-1,-1, 1}, { 0,-2, 0}, {-1,-1,-1} }
	tetras_around_peak[21].top.a = 0.0;
	tetras_around_peak[21].top.b = 0.0;
	tetras_around_peak[21].top.c = 0.0;
	tetras_around_peak[21].second.a = -1.0;
	tetras_around_peak[21].second.b = -1.0;
	tetras_around_peak[21].second.c = 1.0;
	tetras_around_peak[21].third.a = 0.0;
	tetras_around_peak[21].third.b = -2.0;
	tetras_around_peak[21].third.c = 0.0;
	tetras_around_peak[21].bottom.a = -1.0;
	tetras_around_peak[21].bottom.b = -1.0;
	tetras_around_peak[21].bottom.c = -1.0;
//	p[wyz] -->==> { {0,0,0}, { 1,-1,-1}, { 0, 0,-2}, {-1,-1,-1} }
	tetras_around_peak[22].top.a = 0.0;
	tetras_around_peak[22].top.b = 0.0;
	tetras_around_peak[22].top.c = 0.0;
	tetras_around_peak[22].second.a = 1.0;
	tetras_around_peak[22].second.b = -1.0;
	tetras_around_peak[22].second.c = -1.0;
	tetras_around_peak[22].third.a = 0.0;
	tetras_around_peak[22].third.b = 0.0;
	tetras_around_peak[22].third.c = -2.0;
	tetras_around_peak[22].bottom.a = -1.0;
	tetras_around_peak[22].bottom.b = -1.0;
	tetras_around_peak[22].bottom.c = -1.0;
//	p[wzy] -->==> { {0,0,0}, { 1,-1,-1}, { 0,-2, 0}, {-1,-1,-1} }
	tetras_around_peak[23].top.a = 0.0;
	tetras_around_peak[23].top.b = 0.0;
	tetras_around_peak[23].top.c = 0.0;
	tetras_around_peak[23].second.a = 1.0;
	tetras_around_peak[23].second.b = -1.0;
	tetras_around_peak[23].second.c = -1.0;
	tetras_around_peak[23].third.a = 0.0;
	tetras_around_peak[23].third.b = -2.0;
	tetras_around_peak[23].third.c = 0.0;
	tetras_around_peak[23].bottom.a = -1.0;
	tetras_around_peak[23].bottom.b = -1.0;
	tetras_around_peak[23].bottom.c = -1.0;
	return;
}


//===============================================
// Public methods
//===============================================

void tetrahedrons_around(struct hnv_coor_4d * cur_coor)
{
	int i;
	struct hnv_coor_3d cur_top_vertex;
	
	tetrahedrons_around_origin();
	
	// Projection {k, l, m, n} ==> {k-l-m+n, k+l-m-n, k-l+m-n}
	//     x: {1,0,0,0} ==> { 1,  1,  1}
	//     y: {0,1,0,0} ==> {-1,  1, -1}
	//     z: {0,0,1,0} ==> {-1, -1,  1}
	//     w: {0,0,0,1} ==> { 1, -1, -1}
	cur_top_vertex.a = cur_coor->k - cur_coor->l - cur_coor->m + cur_coor->n;
	cur_top_vertex.b = cur_coor->k + cur_coor->l - cur_coor->m - cur_coor->n;
	cur_top_vertex.c = cur_coor->k - cur_coor->l + cur_coor->m - cur_coor->n;
	
	for (i = 0; i < 24; i++) {
		tetras_around_peak[i].top.a += cur_top_vertex.a;
		tetras_around_peak[i].top.b += cur_top_vertex.b;
		tetras_around_peak[i].top.c += cur_top_vertex.c;
		tetras_around_peak[i].second.a += cur_top_vertex.a;
		tetras_around_peak[i].second.b += cur_top_vertex.b;
		tetras_around_peak[i].second.c += cur_top_vertex.c;
		tetras_around_peak[i].third.a += cur_top_vertex.a;
		tetras_around_peak[i].third.b += cur_top_vertex.b;
		tetras_around_peak[i].third.c += cur_top_vertex.c;
		tetras_around_peak[i].bottom.a += cur_top_vertex.a;
		tetras_around_peak[i].bottom.b += cur_top_vertex.b;
		tetras_around_peak[i].bottom.c += cur_top_vertex.c;
	}
}


void set_initial_tetra_around_peak(struct tetrahedron * ft, int i)
{
	set_tetra_vertex(ft, &(tetras_around_peak[i].top), 
			 &(tetras_around_peak[i].second), 
			 &(tetras_around_peak[i].third),
			 &(tetras_around_peak[i].bottom));
	set_tetra_locator(ft);
	return;
}
